/*  character-set based brute-forcing algorithm
 *  Copyright (C) 2004 by mimayin@aciiid.ath.cx
 *  Copyright (C) 2009, 2010 by bofh28@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 only of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  crunch version 1.0 was written by mimayin@aciiid.ath.cx
 *  all later versions of crunch have been updated by bofh28@gmail.com
 *
 *  Changelog:
 *  version 1.0 initial release in 2004 (guess based on copyright date) 194 lines of code
 *  version 1.1 added support to read character list from a specified file
 *  version 1.2 added output parameter
 *  version 1.3 added -c parameter to break apart output file. I would prefer to break
 *              apart on file size instead of number of lines, but there is no reliable
 *              way to get that information while the file is being created.
 *  version 1.4 added better help documentation and fixed some bugs:
 *              space character was sometimes getting stripped off in charset
 *              replace if(pattern[0]==0) { with if((strncmp(&pattern[0],"\0",1)==0)) {
 *              to make splint happy
 *              rename templ variable to pattern
 *              changed some fixed size local buffers to dynamically allocated buffers
 *              if max-len was larger than strlen of the -t parameter crunch would generate
 *              duplicate words
 *              documented the code
 *  version 1.5 fixed length and -s parameter.  Before if you did ./crunch 1 3 -s j
 *              you would only get j-z and not j-zzz
 *              converted some fixed length buffers to variable length
 *              added checks to fclose
 *              fixed it so that -s and -t can now work together
 *              added support to generate wordlists greater than 2GB even though several
 *              programs can't access files of this size (john the ripper for example)
 *              This support works on Ubuntu intrepid.  This probably works on most Linux
 *              platforms.  Support for Solaris, HP-UX, Windows, BSD is unknown.
 *  version 1.6 add option (-p) to support factorial instead of exponentail processing.  i.e.
 *              if you did ./crunch 3 3 abc you would get 27 results (3x3x3): a through ccc
 *              This new option will give you 6 results (3! = 3x2x1)
 *              i.e. (abc, acb, bac, bca, cab, cba)
 *              the permute function was written by Richard Heathfield
 *              and copied from
 *              http://bytes.com/groups/c/536779-richard-heathfields-tsp-permutation-algorithm
 *              I did change temp from an int to char to make splint happy.
 *              Richard Heathfield may be contacted at rjh@cpax.org.uk
 *
 *  version 1.7 add option (-m) to support factoring words instead of characters
 *
 *  version 1.8 add option (-z) to support compressing the output
 *
 *  version 1.9 only need permute2 (remove permute and renamed permute2 to permute)
 *              fix off by 1 error - reported by MikeCa
 *              fix issue where output file wasn't being renamed
 *              fork compression process
 *              really fix it so that -s and -t can now work together
 *              when using -c sometimes the output was 1 larger than requested
 *              cstring_cmp from  http://www.anyexample.com/programming/c/qsort__sorting_array_of_strings__integers_and_structs.xml to sort words to permute so the results will be in sorted order
 *
 *  version 2.0 permute now supports -c -o & -z. -f is also supported for permuting letters
 *              really really fix it so that -s and -t can now work together
 *              added null checks for parameter values to prevent segmentation faults
 *              added support to invert chunk output - written by Pillenski
 *              added support to breakup file based on size - written by Pillenski
 *
 *  version 2.1 permute now supports -b
 *              add signal handling - pressing ctrl break will cause crunch
 *              to finish the word it is processing and then gracefully exit instead of
 *              leaving files open and words half finished.  This is the first step in
 *              supporting resume.
 *              chunk now supports resume
 *
 *  version 2.2 pattern supports numbers and symbols
 *
 *  version 2.3 fix bytecount
 *              add new options to usage
 *
 *  version 2.4 fix usage (-m and not -n) - reported by Tape
 *              clarified -b option in help and man file - reported by Tape
 *              fix Makefile to install to /pentest/passwords instead of /pentest/password -
 *              reported by Tape
 *              make bytecount behave consistently - reported by Tape
 *              fixed up copyrights
 *
 *  TODO:  add resume support to permute (I am not sure this is possible)
 *         make permute more intelligent (min, max) (I am not sure this is possible either)
 *         make -m support reading from strings from file
 *
 *  usage: ./crunch <min-len> <max-len> [charset] [-t [FIXED]@@@@] [-s startblock]
 *  e.g: ./crunch 3 7 abcdef
 *
 *  This example will compute all passwords between 3 and 7 chars
 *  using 'abcdef' as the character set and dump it to stdout.
 *
 *  usage: ./crunch <from-len> <to-len> [-f <path to charset.lst> charset-name] [-o wordlist.txt or START] [-t [FIXED]@@@@] [-s startblock]
 *
 *  Options:
 *  -t [FIXED]@%^   :  allows you to specify a pattern, eg: @@god@@@@ 
 *                     where the only the @'s will change with letters
 *                     the %'s will change with numbers
 *                     the ^'s will change with symbols
 *  -s              :  allows you to specify the starting string, eg: 03god22fs
 *  -o              :  allows you to specify the file to write the output to, eg:
 *                     wordlist.txt
 *  -f              :  path to a file containing a list of character sets, eg: charset.lst
 *                     name of the character set in the above file eg:
 *                     mixalpha-numeric-all-space
 *  -c              :  numbers of lines to write to output file, only works if "-o START"
 *                     is used, eg: 60  The output files will be in the format of starting
 *                     letter - ending letter for example:
 *                     crunch 1 5 -f /pentest/password/charset.lst mixalpha -o START -c 52
 *                     will result in 2 files: a-7.txt and 8-\ .txt  The reason for the
 *                     slash in the second filename is the ending character is space and
 *                     ls has to escape it to print it.  Yes you will need to put in
 *                     the \ when specifying the filename.
 *  -p              :  prints permutations without repeating characters.  This option
 *                     CANNOT be used with -s, or -t.  It also ignores min and max
 *                     lengths.
 *  -m              :  prints permutations without repeating strings.  This option
 *                     CANNOT be used with -s, or -t.  It also ignores min and max
 *                     lengths.
 *  -z              :  adds support to compress the generated output.  Must be used
 *                     with -o option.  Only supports gzip, bzip, and lzma.
 *  -i              :  inverts the output so the first character will change very often
 *  -b              :  maximum bytes to write to output file. depending on the blocksize
 *                     files may be some bytes smaller than specified but never bigger.
 *  -r              :  resume a previous session.  You must use the same command line as
 *                     the previous session.
 *
 *  This code can be easily adapted for use in brute-force attacks
 *  against network services or cryptography.
 *
 *  Compiles on: linux (32 bit Ubuntu for sure, 32 bit Linux in general should work,
 *     64 bit Linux I am not sure.  Please let me know.
 *  It should compile on ms visual c++ and freebsd and the others but I don't
 *     don't have access to any of the those systems.  Please let me know.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>

static const char def_charset[] = "abcdefghijklmnopqrstuvwxyz";
static const char def_num_charset[] = "0123456789";
static const char def_sym_charset[] = "!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/ ";
static const char version[] = "2.4";
static size_t inc[128];
static unsigned long count=0;   /* user specified count */
static unsigned long counter=0; /* counter */
static unsigned long long r=0;  /* count number of bytes */
static int inverted=0;  /* 0 for normal output 1 for aaa,baa,caa,etc */
static int ctrlbreak=0; /* 0 for user did NOT press Ctrl-C 1 for they did */
static FILE *fptr;      /* file pointer */

static int cstring_cmp(const void *a, const void *b) {
  const char **ia = (const char **)a;
  const char **ib = (const char **)b;
  return strcmp(*ia, *ib);
}

static int finished(const char *block, const char *charset, const char *num_charset, const char *sym_charset, const char *pattern) {
size_t i;

  if (pattern == NULL) {
    for (i=0; i < strlen(block); i++)
      if (inc[i] < (strlen(charset)-1))
        return 0;
  }
  else {
    for (i=0; i < strlen(block); i++) {
      if (((strncmp(&pattern[i],"@",1) == 0) && (inc[i] < (strlen(charset)-1))) || ((strncmp(&pattern[i],"%",1) == 0) && (inc[i] < (strlen(num_charset)-1))) || ((strncmp(&pattern[i],"^",1) == 0) && (inc[i] < (strlen(sym_charset)-1)))) {
        return 0;
      }
    }
  }
  return 1;
}

static void increment(char *block, const char *charset, const char *num_charset, const char *sym_charset, const char *pattern) {
size_t i;
size_t cslen=0;
size_t nlen=0;
size_t slen=0;
size_t blocklen = strlen(block);

  if (charset != NULL) cslen = strlen(charset)-1;
  if (num_charset != NULL) nlen = strlen(num_charset)-1;
  if (sym_charset != NULL) slen = strlen(sym_charset)-1;

  if (inverted == 0) {
    for (i=blocklen-1; (int)i>-1; i--) {
      if (pattern==NULL) {
        if ((inc[i] < cslen)) {
          block[i] = charset[++inc[i]];
          i=-1;
        }
        else {
          block[i] = charset[0];
          inc[i] = 0;
        }
      }
      else {
        switch (pattern[i]) {
          case '@':
            if (inc[i] < cslen) {
              block[i] = charset[++inc[i]];
              i=-1;
            }
            else {
              block[i] = charset[0];
              inc[i] = 0;
            }
            break;
          case '%':
            if (inc[i] < nlen) {
              block[i] = num_charset[++inc[i]];
              i=-1;
            }
            else {
              block[i] = num_charset[0];
              inc[i] = 0;
            }
            break;
          case '^':
            if (inc[i] < slen) {
              block[i] = sym_charset[++inc[i]];
              i=-1;
            }
            else {
              block[i] = sym_charset[0];
              inc[i] = 0;
            }
            break;
          default:
            block[i] = pattern[i]; /* add pattern letter to word */
        }
      }
    }
  }
  else {
    for (i=0; i<blocklen; i++) {
      if (pattern==NULL) {
        if (inc[i] < cslen) {
          block[i] = charset[++inc[i]]; 
          i=blocklen;
        }
        else {
          block[i] = charset[0];
          inc[i] = 0;
        }
      }
      else {
        switch (pattern[i]) {
          case '@':
            if (inc[i] < cslen) {
              block[i] = charset[++inc[i]];
              i=-1;
            }
            else {
              block[i] = charset[0];
              inc[i] = 0;
            }
            break;
          case '%':
            if (inc[i] < nlen) {
              block[i] = num_charset[++inc[i]];
              i=-1;
            }
            else {
              block[i] = num_charset[0];
              inc[i] = 0;
            }
            break;
          case '^':
            if (inc[i] < slen) {
              block[i] = sym_charset[++inc[i]];
              i=-1;
            }
            else {
              block[i] = sym_charset[0];
              inc[i] = 0;
            }
            break;
          default:
            block[i] = pattern[i]; /* add pattern letter to word */
        }
      }
    }
  }
}

static void renamefile(const size_t end, const char *outputfilename, const char *compressalgo) {
FILE *optr;     /* ptr to START output file; will be renamed later */
char *newfile;  /* holds the new filename */
int status;     /* rename returns int */
char buff[512]; /* buffer to hold line from wordlist */
pid_t pid, pidret; /* pid and pid return */

  errno=0;
  memset(buff,0,sizeof(buff));

  newfile = (char*) calloc((end*3)+5,sizeof(char)); /* max length will be 3x outname */
  if (newfile == NULL) {
    printf("rename: can't allocate memory for newfile\n");
    exit(EXIT_FAILURE);
  }

  if (strncmp(outputfilename,"START",5) != 0) {
    status=rename("START",outputfilename); /* rename from START to user specified name */
    if (status != 0) {
      printf("1Error renaming file.  Status = %d  Code = %d\n",status,errno);
      printf("The problem is = %s\n", strerror(errno));
      exit(EXIT_FAILURE);
    }
  }
  else {
    if ((optr = fopen("START","r")) == NULL) {
      printf("rename: File START could not be opened\n");
      exit(EXIT_FAILURE);
    }
    else {  /* file opened above now read first line */
      (void)fgets(buff, (int)sizeof(buff),optr);
      memcpy(newfile,buff,strlen(buff)-1); /* get rid of CR */
      while (feof(optr)==0) {
        (void)fgets(buff, (int)sizeof(buff),optr);
      } /* all of this just to get last line */
      if (fclose(optr) != 0) {
        printf("rename: fclose returned error number = %d\n",errno);
        printf("The problem is = %s\n", strerror(errno));
        exit(EXIT_FAILURE);
      }
    }
    strcat(newfile,"-"); /* build new filename */
    strncat(newfile,buff,strlen(buff)-1); /* get rid of CR */
    strncat(newfile,".txt",4);
    status=rename("START",newfile); /* rename from START to proper filename */
    if (status != 0) {
      printf("2Error renaming file.  Status = %d  Code = %d\n",status,errno);
      printf("The problem is = %s\n", strerror(errno));
      exit(EXIT_FAILURE);
    }
  }

  if (compressalgo != NULL) {
    pid = fork();
    pidret = wait(&status);
    if (pid == 0) {
      if (strncmp(outputfilename,"START",5) == 0)
        status=execlp(compressalgo, compressalgo, "-9", "-f", newfile, NULL);
      else
        status=execlp(compressalgo, compressalgo, "-9", "-f", outputfilename, NULL);
      if (status < 0) {
        printf("Error compressing file.  Status = %d  Code = %d\n",status,errno);
        printf("The problem is = %s\n", strerror(errno));
        exit(EXIT_FAILURE);
      }
    }
  }
  free(newfile);
}

static void ex_program() {
ctrlbreak=1;
(void) signal(SIGINT, SIG_DFL);
}

static void Permute(const char *outputfilename, const char *compressalgo, char *Perm[], const size_t sizePerm, size_t unchanged, const unsigned long long *bytecount) {
size_t outer = 0;
size_t inner = 0;
size_t t;
char *temp[1];

  if(sizePerm > unchanged) {
    for(outer = unchanged; outer < sizePerm; outer++) {
      *temp = Perm[outer];
      for(inner = outer; inner > unchanged; inner--) {
        Perm[inner] = Perm[inner - 1];
      }
      Perm[unchanged] = *temp;
      Permute(outputfilename,compressalgo, Perm, sizePerm, unchanged+1, bytecount);

      for(inner = unchanged; inner < outer; inner++) {
        Perm[inner] = Perm[inner + 1];
      }
      Perm[outer] = *temp;
    }
  }
  else {
    if (outputfilename != NULL) {
      if ((fptr = fopen("START","a+")) == NULL) { /* create file */
        printf("1permute: File START could not be opened\n");
        printf("The problem is = %s\n", strerror(errno));
        exit(EXIT_FAILURE);
      }
    }
    for (t=0;t<sizePerm;t++) {
      fprintf(fptr, "%s", Perm[t]);
      r+=(unsigned long long)strlen(Perm[t]);
    }
    fprintf(fptr, "\n");
    r++;
    counter++;

    if (outputfilename != NULL) {
      if (fclose(fptr) != 0) {
        printf("permute: fclose returned error number = %d\n",errno);
        printf("The problem is = %s\n", strerror(errno));
        exit(EXIT_FAILURE);
      }
    }

    if ((counter == count) || ((r > *bytecount) && (*bytecount != 0))) {
      renamefile(sizePerm,outputfilename,compressalgo);
      counter = 0;
      r=0;
    }

    if (ctrlbreak == 1)
      exit(EXIT_FAILURE);
  }
}

static void chunk(const size_t start, const size_t end, const char *charset, const char *num_charset, const char *sym_charset, const char *startblock, const char *pattern, const unsigned long long *bytecount, const char *outputfilename, const char *compressalgo) {
size_t i,j,k;      /* loop counters */
char *block;       /* block is word being created */

  errno=0;
  block = (char*) calloc(end+1,sizeof(char)); /* block can't be bigger than max size */
  if (block == NULL) {
    printf("chunk: can't allocate memory for block\n");
    exit(EXIT_FAILURE);
  }
  for (i=start; (i<=end) && (ctrlbreak==0); i++) {
    for (j=0; j<i; j++) {
      if (startblock==NULL) { /* is startblock null? */
        if (pattern==NULL) { /* Yes so now test if pattern null? */
          block[j] = charset[0]; /* pattern is null so add character */
          inc[j] = 0;
        }
        else { /* pattern is not null */
          switch (pattern[j]) {
          case '@':
            block[j] = charset[0]; /* placeholder is set so add character */
            inc[j] = 0;
            break;
          case '%':
            block[j] = num_charset[0]; /* placeholder is set so add character */
            inc[j] = 0;
            break;
          case '^':
            block[j] = sym_charset[0]; /* placeholder is set so add character */
            inc[j] = 0;
            break;
          default:
            block[j] = pattern[j]; /* add pattern letter to word */
          }
        }
      }
      else { /* startblock is not null */
        if (pattern==NULL) { /* Yes so now test if pattern null? */
          block[j] = startblock[j]; /* pattern is null so add character */
          for(k=0;k<strlen(charset);k++)
            if(block[j]==charset[k]) {
              inc[j] = k;
            }
        }
        else {
          if (strncmp(&pattern[j],"@",1) == 0) { /* test for placeholder */
            block[j] = startblock[j]; /* pattern is null so add character */
            for(k=0;k<strlen(charset);k++)
              if(block[j]==charset[k])
                inc[j] = k;
          }
          else {
            block[j] = pattern[j]; /* add pattern letter to word */
          }
        }
      }
    } /* for j=0 */
    startblock = NULL;

    if (outputfilename==NULL) { /* user wants to display words on screen */
      printf("%s\n",block);
      while ((finished(block,charset,num_charset,sym_charset,pattern)==0) && (ctrlbreak==0)) {
        increment(block,charset,num_charset,sym_charset,pattern);
        printf("%s\n",block);
      }
    }
    else { /* user wants to generate a file */
      if ((fptr = fopen("START","a+")) == NULL) { /* append to file */
        printf("chunk1: File START could not be opened\n");
        printf("The problem is = %s\n", strerror(errno));
        exit(EXIT_FAILURE);
      }
      else { /* file opened start writing.  file will be renamed when done */
        while ((finished(block,charset,num_charset,sym_charset,pattern)==0) && (ferror(fptr)==0) && (ctrlbreak==0)) {
          if ((counter <= (count)-1) && (r <= (*bytecount-strlen(block)))) { /* not time to create a new file */
            fprintf(fptr,"%s\n",block);
            if (ferror(fptr) != 0) {
              printf("chunk1: fprintf failed = %d\n", errno);
              printf("The problem is = %s\n", strerror(errno));
              free(block);
              exit(EXIT_FAILURE);
            }
            increment(block,charset,num_charset,sym_charset,pattern);
            r+=(unsigned long long)strlen(block)+1;
            counter++;
          }
          else { /* time to create a new file */
            if (fclose(fptr) != 0) {
              printf("chunk1: fclose returned error number = %d\n",errno);
              printf("The problem is = %s\n", strerror(errno));
              free(block);
              exit(EXIT_FAILURE);
            }
            renamefile(end,outputfilename,compressalgo);
            if((fptr = fopen("START","w")) == NULL) {
              printf("chunk2: Ouput file START could not be opened\n");
              free(block);
              exit(EXIT_FAILURE);
            }
            counter=0;
            r=0;
          }
        }
        fprintf(fptr,"%s\n",block); /* flush the last word */
        r+=(unsigned long long)strlen(block)+1;
        counter++;

        if (ferror(fptr) != 0) {
          printf("chunk2: fprintf failed = %d\n", errno);
          printf("The problem is = %s\n", strerror(errno));
          free(block);
          exit(EXIT_FAILURE);
        }
        if (fclose(fptr) != 0) {
          printf("chunk2: fclose returned error number = %d\n",errno);
          printf("The problem is = %s\n", strerror(errno));
          free(block);
          exit(EXIT_FAILURE);
        }
        if (ctrlbreak == 1)
          break;
      } /* else from if fopen */
    } /* else from outputfilename == NULL */
  } /* for start < end loop */
  if ((outputfilename != NULL) && (ctrlbreak == 0)) {
    renamefile(end,outputfilename,compressalgo);
  }

  free(block);
}

static void usage() {
  printf("crunch version %s\n\n", version);
  printf("usage: crunch <min-len> <max-len> [charset] [-o wordlist.txt] [-t [FIXED]@@@@] [-s startblock] [-c number]\n");
  printf("or\n");
  printf("usage: crunch <min-len> <max-len> [-f <path to charset.lst> charset-name] [-o wordlist.txt] [-t [FIXED]@@@@] [-s startblock] [-c number]\n");
  printf("\n");
  printf("min-len is the minimum length string you want crunch to start at\n");
  printf("max-len is the maximum length string you want crunch to end at\n\n");
  printf("[charset] is optional.  You may specify a character set for crunch to use on the command line or if you leave it blank crunch will use abcdefghijklmnopqrstuvwxyz as the character set.  NOTE: If you want to include the space character in your character set you use enclose your character set in quotes i.e. \"abc \"\n\n");
  printf("[-f <path to charset.lst> <charset-name>] is the alternative to setting the character set on command line.  This parameter allows you to specify a character set from the charset.lst.\n");
  printf("NOTE: You may either not specify a charset, you may specify a character set on the command line, or you may specify -f <path to charset.lst> <charset-name>.  You can only do one.\n\n");
  printf("[-t [FIXED]@%%^] is optional and allows you to specify a pattern, eg: @@god@@@@ where the only the @'s will change.  You can specify @ or %% or ^  which will allow to change letters numbers and symbols.  For example @@dog%%^ abcd 1234 @#$%% the @'s will chnage with the characters the %% will change with the numbers and the ^ will change with the symbols you specifed.  If you use %%dog^ + 1234 @#$%% the plus sign is a placeholder.\n\n");
  printf("[-s startblock] is optional and allows you to specify the starting string, eg: 03god22fs\n\n");
  printf("[-i] is optional and will invert the output.  Instad of aaa, aab, aac you will get aaa, baa, caa, etc\n\n");
  printf("[-o wordlist.txt] is optional allows you to specify the file to write the output to, eg: wordlist.txt\n\n");
  printf("[-c number] is optional and specifies the number of lines to write to output file, only works if -o START is used, eg: 60  The ouput files will be in the format of starting letter-ending letter for example:\n ./crunch 1 1 -f /pentest/password/crunch/charset.lst\n mixalpha-numeric-all-space -o START -c 60\n will result in 2 files: a-7.txt and 8-\\ .txt  The reason for the slash in  the second filename is the ending character is space and ls has to escape it to print it.  Yes you will need to put in the \\ when specifing the filename.\n\n");
  printf("[-b size] is optional and will break apart the output file based on the size you specify.  you must specify the size as a number followed by kb, mb, gb, kib, mib, or gib.  kb, mb, and gb are based on the power of 10 (i.e. 1 kb = 1000 bytes).  kib, mib, and gib are based on the power of two (i.e. 1 kb = 1024 bytes).  To use this option you must specify -o START.  NOTE There is no space between the number and the type.  For example 500mb is correct, 500 mb is NOT correct.\n\n");
  printf("[-z option] is optional and will compress the outfile using bzip, gzip, or lzma\n\n");
  printf("[-p string-of-letters] is optional crunch will now generate permutations instead of combinations.  The string of letters is the string of letters you use to permute.  This parameter ignores min and max\n\n");
  printf("[-m string1 string2 string3 ...] is optional crunch will now generate permututions of the strings you specify.  This parameter ignores min and max\n\n");
  printf("[-r] is optional and resume a previous crunch session.  You must use the same syntax as the original command.  This option will not work with the -p or -m options.\n\n");
  printf("examples:\n");
  printf("./crunch 1 8\n");
  printf("crunch will display a wordlist that starts at a and ends at zzzzzzzz\n\n");
  printf("./crunch 1 6 abcdefg\n");
  printf("crunch will display a wordlist using the charcterset abcdefg that starts at a and ends at gggggg\n\n");
  printf("./crunch 1 8 -f charset.lst mixalpha-numeric-all-space -o wordlist.txt\n");
  printf("crunch will use the mixalpha-numberic-all-space character set from charset.lst and will write the wordlist to a file named wordlist.txt.  The file will start with a and end with \"        \"\n\n");
  printf("./crunch 8 8 -f charset.lst mixalpha-numeric-all-space -o wordlist.txt -t @@dog@@@ -s cbdogaaa \n");
  printf("crunch will generate a 8 character wordlist using the mixalpha-number-all-space characterset from charset.lst and will write the wordlist to a file named wordlist.txt.  The file will start at cbdogaaa and end at \"  dog   \"\n");
}

static char *resumesession(const char *charset) {
FILE *optr;     /* ptr to START output file; will be renamed later */
char buff[512]; /* buffer to hold line from wordlist */
char *startblock;
size_t j,k;

  errno=0;
  memset(buff,0,sizeof(buff));

  if ((optr = fopen("START","r")) == NULL) {
    printf("resume: File START could not be opened\n");
    exit(EXIT_FAILURE);
  }
  else {  /* file opened above now read first line */
    while (feof(optr)==0) {
      (void)fgets(buff, (int)sizeof(buff),optr);
      counter++;
    } /* all of this just to get last line */
    if (fclose(optr) != 0) {
      printf("resume: fclose returned error number = %d\n",errno);
      printf("The problem is = %s\n", strerror(errno));
      exit(EXIT_FAILURE);
    }
    startblock = (char*) calloc(strlen(buff),sizeof(char));
    if (startblock == NULL) {
      printf("crunch: can't allocate memory for startblock\n");
      exit(EXIT_FAILURE);
    }
    memcpy(startblock,buff,strlen(buff)-1);
  }
  printf("Resuming from = %s\n",startblock);
  for (j=0; j<strlen(startblock);j++) {
    for(k=0;k<strlen(charset);k++)
      if(startblock[j]==charset[k])
        inc[j] = k;
  }

  return startblock;
}

static char *readcharsetfile(const char *charfilename, const char *charsetname) {
FILE *optr;  /* ptr to user specified charset file */
char *temp;  /* holds character set name from charset file */
char *chars; /* holds characters from specified charsetname i.e. stuff after [ */
char *charset=NULL; /* holds the characters to used */
char buff[512]; /* buffer to hold line from charset file */
errno=0;

  memset(buff,0,sizeof(buff));
  if ((optr = fopen(charfilename,"r")) == NULL) { /* open file to read from */
    printf("charset: File %s could not be opened\n", charfilename);
    printf("The problem is = %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  else {  /* file opened above now scan file for charsetname */
    while (fgets(buff, (int)sizeof(buff),optr) != NULL) {
      temp = strtok(buff, " ");
      if (strcmp(charsetname,temp)==0) {
        chars = strtok(NULL, "["); /* move to = sign in file */
        if ((chars = strtok(NULL, "\n")) != NULL) { /* get rest of string after [ */
          charset = (char*) calloc(strlen(chars),sizeof(char));
          if (charset == NULL) {
            printf("charset: can't allocate memory for charset\n");
            exit(EXIT_FAILURE);
          }
          if (strncmp(&chars[(strlen(chars)-1)],"]",1) == 0)
            memcpy(charset,chars,strlen(chars)-1); /* don't strip off space only ]*/
          else
            memcpy(charset,chars,strlen(chars)-2); /* strip off ] */
        }
        break;
      }
    }
    if (fclose(optr) != 0) {
      printf("charset: fclose returned error number = %d\n",errno);
      printf("The problem is = %s\n", strerror(errno));
      exit(EXIT_FAILURE);
    }
  }
  return charset;
}

int main(int argc, char *argv[]) {
int i=3;      /* minimum number of parameters */
int flag=0;   /* 0 for chunk 1 for permute */
int loaded=0; /* loaded wordarray from specified file */
int resume=0; /* 0 new session 1 for resume */

size_t arglen=0;
size_t numleft=0; /* number of arguments left to read from command line */
size_t min,max;   /* minimum and maximum size */
size_t temp;      /* temp varible */
size_t patterncount=0;

unsigned long calc=0;  /* recommend count */
unsigned long long bytecount=0;
int multi=0;

char *charset; /* character set */
char *num_charset=NULL;
char *sym_charset=NULL;
char *startblock=NULL; /* user specified starting point */
char *pattern=NULL; /* user specified pattern */
char *outputfilename=NULL; /* user specified filename to write output to */
char *compressalgo=NULL; /* user specified compression program */

char *bcountval=NULL;
char **wordarray; /* array to store words */

  (void) signal(SIGINT, ex_program);

  fptr=stdout;
  if (argc<3) {
    usage();
    return 0;
  }

  if ((argc >= 4) && (strncmp(argv[3],"-",1) != 0)) { /*test for ./crunch 1 2 -? */
    charset = (char*) calloc(strlen(argv[3])+1,sizeof(char)); /* user specified charset */
    if (charset == NULL) {
      printf("crunch: can't allocate memory for user charset\n");
      return -1;
    }
    /* remove duplicate characters from user defined charset */
    for (min=0; min<strlen(argv[3]); min++) {
      temp = 0;
      for (max=0; max<strlen(charset); max++)
        if (argv[3][min] == charset[max])
          temp = 1;
        if (temp==0)
          strncat(charset,&argv[3][min],1);
    }
    i=4; /* user specified a charaset and a parameter */
  }
  else { /* user didn't specify a charset on command line so use default */
    charset = (char*) calloc(26,sizeof(char));
    if (charset == NULL) {
      printf("crunch: can't allocate memory for default charset\n");
      return -1;
    }
    memcpy(charset,def_charset,26);
  }

  min = (size_t)atoi(argv[1]);
  max = (size_t)atoi(argv[2]);

  if (max < min) {
    printf("Starting length is greater than the ending length\n");
    free(charset);
    return -1;
  }

  if (max > 128) {
    printf("Crunch can only make words with a length of less than 129 characters\n");
    free(charset);
    return -1;
  }

  for (; i<argc; i+=2) { /* add 2 to skip the parameter value */
    if (strncmp(argv[i],"-f",2) == 0) { /* user specified a charset.lst */
      if ((argv[i+1] != NULL) && (argv[i+2] != NULL)) {
        free(charset);
        charset=readcharsetfile(argv[i+1], argv[i+2]);
        if (charset == NULL) {
          printf("charset = %s not found in %s\n", argv[i+2], argv[i+1]);
          exit(EXIT_FAILURE);
        }
        else {
          numleft = strlen(charset);
          wordarray=malloc(numleft * sizeof(*wordarray));
          if (wordarray == NULL) {
            printf("can't allocate memory for wordarray1\n");
            exit(EXIT_FAILURE);
          }
          for (temp=0; temp < numleft; temp++) {
            wordarray[temp]=malloc(2 * sizeof(char)); /* space for letter + \0 char */
            if (wordarray[temp] == NULL) {
              printf("can't allocate memory for wordarray2\n");
              exit(EXIT_FAILURE);
            }
            strncpy(&wordarray[temp][0], &charset[temp], 1);
            strncpy(&wordarray[temp][1], "\0", 1);
          }
          loaded=1;
        }
        i++; /* have to add 1 more since -f has two parameter values */
      }
      else {
        printf("Please specify a filename and character set\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-p",2) == 0) { /* user specified letters to permute */
      if ((argv[i+1] != NULL) || (loaded == 1)) {
        flag=1;
        if (loaded == 0) {
          numleft=strlen(argv[i+1]);
          wordarray=malloc(numleft * sizeof(*wordarray));
          if (wordarray == NULL) {
            printf("can't allocate memory for wordarray1\n");
            exit(EXIT_FAILURE);
          }
          for (temp=0; temp < numleft; temp++) {
            wordarray[temp]=malloc(2 * sizeof(char)); /* space for letter + \0 char */
            if (wordarray[temp] == NULL) {
              printf("can't allocate memory for wordarray2\n");
              exit(EXIT_FAILURE);
            }
            strncpy(&wordarray[temp][0], &argv[i+1][temp], 1);
            strncpy(&wordarray[temp][1], "\0", 1);
          }
        }
      }
      else {
        printf("Please specify letters to permute\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-m",2) == 0) { /* user specified words to permute */
      if (argv[i+1] != NULL) {
        flag=1;
        numleft=(size_t)(argc-i)-1;
        if (numleft > 0) {
          wordarray=malloc(numleft * sizeof(*wordarray));
          if (wordarray == NULL) {
            printf("can't allocate memory for wordarray3\n");
            exit(EXIT_FAILURE);
          }
          wordarray[0]=""; /* too make splint happy */
          for (temp=0; temp < numleft; temp++,i++) {
            wordarray[temp] = argv[i+1];
          }
          /* sort wordarray so the results are sorted */
          qsort(wordarray, temp, sizeof(char *), cstring_cmp);
        }
      }
      else {
        printf("Please specify a words to permute\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-t",2) == 0) { /* pattern specified */
      if (argv[i+1] != NULL) {
        if ((max > strlen(argv[i+1])) || (min < strlen(argv[i+1]))) {
          printf("The maximum and minimum length should be the same size as the pattern you specified. \n");
          exit(EXIT_FAILURE);
        }
        pattern = (char*) calloc(strlen(argv[i+1])+1,sizeof(char));
        if (pattern == NULL) {
          printf("crunch: can't allocate memory for pattern\n");
          exit(EXIT_FAILURE);
        }
        memcpy(pattern, argv[i+1], strlen(argv[i+1]));
        if (strstr(pattern,"@") !=NULL) patterncount++;
        if (strstr(pattern,"%") !=NULL) patterncount++;
        if (strstr(pattern,"^") !=NULL) patterncount++;

        if ((argv[i+2] == NULL) || strncmp(&argv[i+2][0],"-",1)==0){
          num_charset=(char*) calloc(strlen(def_num_charset),sizeof(char));
          if (num_charset == NULL) {
            printf("can't allocate memory for num_charset\n");
            exit(EXIT_FAILURE);
          }
          memcpy(num_charset,def_num_charset,strlen(def_num_charset));
          sym_charset=(char*) calloc(strlen(def_sym_charset),sizeof(char));
          if (sym_charset == NULL) {
            printf("can't allocate memory for sym_charset\n");
            exit(EXIT_FAILURE);
          }
          memcpy(sym_charset,def_sym_charset,strlen(def_sym_charset));
          patterncount=0;
        }
        else {
          if (strstr(pattern,"@") !=NULL) {
            if ((argv[i+2] == NULL) || (strncmp(&argv[i+2][0],"+",1)==0) || (strncmp(&argv[i+2][0],"-",1)==0)){
              free(charset);
              charset=(char*) calloc(strlen(def_charset),sizeof(char));
              if (charset == NULL) {
                printf("can't allocate memory for charset\n");
                exit(EXIT_FAILURE);
              }
              memcpy(charset,def_charset,strlen(def_charset));
            }
            else {
              free(charset);
              charset=(char*) calloc(strlen(argv[i+2])+1,sizeof(char));
              if (charset == NULL) {
                printf("can't allocate memory for charset\n");
                exit(EXIT_FAILURE);
              }
              memcpy(charset,argv[i+2],strlen(argv[i+2]));
            }
          }
          if (strstr(pattern,"%") !=NULL) {
            if ((argv[i+3] == NULL) || (strncmp(&argv[i+3][0],"+",1)==0) || (strncmp(&argv[i+3][0],"-",1)==0)){
              num_charset=(char*) calloc(strlen(def_num_charset),sizeof(char));
              if (num_charset == NULL) {
                printf("can't allocate memory for num_charset\n");
                exit(EXIT_FAILURE);
              }
              memcpy(num_charset,def_num_charset,strlen(def_num_charset));
              patterncount--;
            }
            else {
              free(num_charset);
              num_charset=(char*) calloc(strlen(argv[i+3])+1,sizeof(char));
              if (num_charset == NULL) {
                printf("can't allocate memory for num_charset\n");
                exit(EXIT_FAILURE);
              }
              memcpy(num_charset,argv[i+3],strlen(argv[i+3]));
            }
          }
          if (strstr(pattern,"^") !=NULL) {
            if ((argv[i+4] == NULL) || (strncmp(&argv[i+4][0],"+",1)==0) || (strncmp(&argv[i+4][0],"-",1)==0)){
              sym_charset=(char*) calloc(strlen(def_sym_charset),sizeof(char));
              if (sym_charset == NULL) {
                printf("can't allocate memory for sym_charset\n");
                exit(EXIT_FAILURE);
              }
              memcpy(sym_charset,def_sym_charset,strlen(def_sym_charset));
              patterncount--;
            }
            else {
              free(sym_charset);
              sym_charset=(char*) calloc(strlen(argv[i+4])+1,sizeof(char));
              if (sym_charset == NULL) {
                printf("can't allocate memory for sym_charset\n");
                exit(EXIT_FAILURE);
              }
              memcpy(sym_charset,argv[i+4],strlen(argv[i+4]));
            }
          }
        }
        i+=patterncount; /* skip the extra pattern values */
      }
      else {
        printf("Please specify a pattern\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-o",2) == 0) {  /* outputfilename specified */
      if (argv[i+1] != NULL) {
        outputfilename = (char*) calloc(strlen(argv[i+1])+1,sizeof(char));
        if (outputfilename == NULL) {
          printf("crunch: can't allocate memory for outputfilename\n");
          exit(EXIT_FAILURE);
        }
        memcpy(outputfilename, argv[i+1], strlen(argv[i+1]));
      }
      else {
        printf("Please specify a output filename\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-z",2) == 0) {  /* compression algorithm specified */
      if (argv[i+1] != NULL) {
        compressalgo = (char*) calloc(strlen(argv[i+1])+1,sizeof(char));
        if (compressalgo == NULL) {
          printf("crunch: can't allocate memory for compression algorithm\n");
          exit(EXIT_FAILURE);
        }
        memcpy(compressalgo, argv[i+1], strlen(argv[i+1]));
        if ((strcmp(compressalgo,"gzip") != 0) && (strcmp(compressalgo,"bzip2") != 0) &&
          (strcmp(compressalgo,"lzma") != 0)) {
          printf("Only gzip, bzip2, and lzma are supported\n");
          exit(EXIT_FAILURE);
        }
      }
      else {
        printf("Only gzip, bzip2, and lzma are supported\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-c",2) == 0) {
      if (argv[i+1] != NULL) {
        count=strtoul(argv[i+1],NULL,10);
        if ((count*max) > 2147483648UL) {
          calc = (2147483648UL/max);
          printf("WARNING: resulting file will probably be larger than 2GB \n");
          printf("Some applications (john the ripper) can't use wordlists greater than 2GB\n");
          printf("A value of %lu ",calc);
          printf("or less should result in a file less than 2GB\n");
          printf("The above value is calcualated based on 2147483648UL/max\n");
          exit(EXIT_FAILURE);
        }
      }
      else {
        printf("Please specify the number of lines you want\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-s",2) == 0) { /* startblock specified */
      if (argv[i+1] != NULL) {
        if (strlen(argv[i+1]) != min) {
          printf("Warning: minimum length should be %d\n",(int)strlen(argv[i+1]));
          exit(EXIT_FAILURE);
        }
        startblock = (char*) calloc(strlen(argv[i+1]),sizeof(char));
        if (startblock == NULL) {
          printf("crunch: can't allocate memory for startblock\n");
          exit(EXIT_FAILURE);
        }
        memcpy(startblock,argv[i+1],strlen(argv[i+1]));
      }
      else {
        printf("Please specify the number of lines you want\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-i",2) == 0) { /* user wants to invert output calculation */
      inverted=1;
      i--; /* decrease by 1 since -i has no parameter value */
    }

    if (strncmp(argv[i],"-b",2) == 0) { /* user wants to split files by size */
      if (argv[i+1] != NULL) {
        bcountval = (char*) calloc(strlen(argv[i+1]),sizeof(char));
        if (bcountval == NULL) {
          printf("crunch: can't allocate memory for bcountval\n");
          exit(EXIT_FAILURE);
        }
        memcpy(bcountval,argv[i+1],strlen(argv[i+1]));
        arglen = strlen(bcountval);

        for (temp=0;temp<arglen;temp++)
          bcountval[temp]=tolower(bcountval[temp]);

        if (strstr(bcountval, "kb") != 0) multi=1000;
        else if (strstr(bcountval, "mb")  != 0) multi=1000000;
        else if (strstr(bcountval, "gb")  != 0) multi=1000000000;
        else if (strstr(bcountval, "kib") != 0) multi=1024;
        else if (strstr(bcountval, "mib") != 0) multi=1048576;
        else if (strstr(bcountval, "gib") != 0) multi=1073741824;

        calc = strtoul(bcountval,NULL,10);
        bytecount=calc*multi;
        if (calc>4 && multi>=1073741824 && bytecount<=4294967295ULL) {
          printf("ERROR: Your system is unable to process numbers greater than 4.294.967.295. Please specify a filesize <= 4GiB.\n");
          exit(EXIT_FAILURE);
        }
      }
      else {
        printf("Please specify a value\n");
        exit(EXIT_FAILURE);
      }
    }

    if (strncmp(argv[i],"-r",2) == 0) { /* user wants to resume a previous session */
      resume=1;
      i--; /* decrease by 1 since -r has no parameter value */
    }
  } /* end parameter processing */

  if (outputfilename != NULL) {
    if (((count > 0) || (bytecount > 0)) && (strcmp(outputfilename,"START") != 0)) {
      printf("you must use -o START if you specify a count\n");
      exit(EXIT_FAILURE);
    }
  }

  if (resume == 1) {
    if (startblock != NULL) {
      printf("you cannot specify a startblock and resume\n");
      exit(EXIT_FAILURE);
    }
    if (flag == 0) {
      startblock=resumesession(charset);
      min = strlen(startblock);
      increment(startblock,charset,num_charset,sym_charset,pattern); /* prevent duplicate words in output */
    }
    if (flag == 1) {
      printf("permute doesn't support resume\n");
      exit(EXIT_FAILURE);
    }
  }
  else {
    (void)remove("START");
  }

  if (flag == 0)
    chunk(min,max,charset,num_charset,sym_charset,startblock,pattern,&bytecount,outputfilename,compressalgo);
  else {
    Permute(outputfilename, compressalgo, wordarray, numleft, 0, &bytecount);
    free(wordarray);
    if ((outputfilename != NULL) && (counter !=0))
      renamefile(numleft,outputfilename,compressalgo);
  }

  free(pattern);
  free(startblock);
  free(charset);
  free(num_charset);
  free(sym_charset);
  free(outputfilename);
  free(compressalgo);
  free(bcountval);
  return 0;
}
